# -*- coding: utf-8 -*-

__author__ = "Ronald Bister, Mike Boutillier"
__credits__ = ["Ronald Bister", "Mike Boutillier"]
__maintainer__ = "Ronald Bister"
__email__ = "mini.pelle@gmail.com"
__license__ = "Apache 2.0"
__version__ = "0.7.2"
